
path = 'SyntheticData\Test_ConsistentLight\00000.';

%load image, depth, vertex, normal, data, flow and computes the point cloud
C = cell(2);
C{1} = loadData(path, 0, '');
C{2} = loadData(path, 1, '');

%warp images using the optical flow
C{1}.image_w = warp(C{2}.image, C{1}.flow_mask, C{1}.flowX, C{1}.flowY);
C{2}.image_w = warp(C{1}.image, C{2}.flow_mask, C{2}.flowX, C{2}.flowY);

%display the optical flows and the warped images
subplot(2,3,1);
imagesc((C{1}.flow)); axis image off; title('Flow01');
subplot(2,3,2);
imagesc(uint8(C{1}.image)); axis image off; title('Image0');
subplot(2,3,3);
imagesc(uint8(C{1}.image_w)); axis image off; title('warped Image1');
subplot(2,3,4);
imagesc((C{2}.flow)); axis image off; title('Flow10');
subplot(2,3,5);
imagesc(uint8(C{2}.image)); axis image off; title('Image1');
subplot(2,3,6);
imagesc(uint8(C{2}.image_w)); axis image off; title('warped Image1');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function C = loadData(path, i, net)
    % loads all avaliable data and returns an object with all structures 

    %load image
    C.image = imread([path, 'image', num2str(i), '.png']);
    C.image = cat(3, C.image, C.image, C.image);

    %load data, containing calibration, min/max depths and min/max flows
    fid = fopen([path, 'data', num2str(i), net, '.json'], 'r'); 
    str = char(fread(fid,inf)'); 
    fclose(fid); 
    C.data = jsondecode(str);

    %load depth map and compute mask and vertex map
    C.depth = double(imread([path, 'depth', num2str(i), '.png']));
    C.mask = C.depth~=0;
    C.depth = C.mask .* (C.depth*(C.data.maxDepth-C.data.minDepth)/65535 + C.data.minDepth);
    C.vertex = depth2vertex(C.depth, C.data.K);
    
    %load normals and transform with respect to camera rotation
    C.normal = double(imread([path, 'normal', num2str(i), '.png']));
    C.normal = cat(3, C.mask, C.mask, -C.mask) .* (cat(3, C.normal(:,:,1), C.normal(:,:,2), C.normal(:,:,3))/255*2 - 1);
    C.normal = C.data.R' * reshape(C.normal, [], 3)';
    C.normal = double(C.mask) .* reshape(C.normal', size(C.vertex));
    
    %compute point clouds
    camera = camera_visualization;
    camera(1:3,:) = C.data.R'*camera(1:3,:)/4 - repmat(C.data.R'*C.data.t,1,size(camera,2));
    light = light_visualization;
    light(1:3,:) = light(1:3,:)/8 + repmat(C.data.lightPos,1,size(light,2));
    
    C.pointCloud = RGBDN2pointCloud(C.image, C.depth, C.normal, C.data);
    
    %savePLY([path, 'pointcloud', num2str(i), net, '.ply'], C.pointCloud);
    savePLY([path, 'pointcloudCalib', num2str(i), net, '.ply'], [C.pointCloud, camera, light]);

    %load optical flow
    flow = double(imread([path, 'flow', num2str(i), '.png']));
    C.flow_mask = prod(flow(:,:,1:2)~=0,3);
    C.flowX = C.flow_mask .* (flow(:,:,1)/65535*(C.data.maxFlowX-C.data.minFlowX) + C.data.minFlowX);
    C.flowY = C.flow_mask .* (flow(:,:,2)/65535*(C.data.maxFlowY-C.data.minFlowY) + C.data.minFlowY);
    flowMag = sqrt(max(abs(C.data.maxFlowX), abs(C.data.minFlowX)).^2 + max(abs(C.data.maxFlowY), abs(C.data.minFlowY)).^2);
    C.flow = flow2rgb(C.flowX, C.flowY, flowMag);
    %imwrite(C.flow, [path, 'flowDir', num2str(i), '.png']);
end

function vertex = depth2vertex(depth, K)
    % creates vertex map from depth map and calibration
    
    [X,Y] = meshgrid(1:size(depth,2), 1:size(depth,1));
    X = X - K(1,3);
    Y = Y - K(2,3);
    Z = ones(size(depth)) * K(1,1);
    Dir = cat(3,X,Y,Z);
    vertex = Dir .* repmat(depth ./ sqrt(sum(Dir.^2,3)),1,1,3);
end

function pointCloud = RGBDN2pointCloud(im, depth, normal, P)
    % creates point cloud array consisting of vertices, normals and colors
    
    [h, w] = size(depth);
    vertex = depth2vertex(depth, P.K);
    vertex = repmat(-P.R' * P.t,1,h*w) + P.R' * reshape(vertex, [], 3)';
    vertex = double(depth~=0) .* reshape(vertex', h, w, 3);

    vertices = reshape(vertex, [], 3)';
    normals = reshape(normal, [], 3)';
    colors = reshape(im, [], 3)';
    
    pointCloud = [vertices(:,depth(:)~=0); normals(:,depth(:)~=0); double(colors(:,depth(:)~=0))];
end

function points = camera_visualization(  )
    % creates a point cloud visualization of a small camera

    points1 = [ 0.5*ones(1,101);  0.5*ones(1,101); -0.5:0.01:0.5];
    points2 = [-0.5*ones(1,101);  0.5*ones(1,101); -0.5:0.01:0.5];
    points3 = [ 0.5*ones(1,101); -0.5*ones(1,101); -0.5:0.01:0.5];
    points4 = [-0.5*ones(1,101); -0.5*ones(1,101); -0.5:0.01:0.5];
    points = [points1, points2, points3, points4];
    points = [points, points([1,3,2],:), points([2,1,3],:), ...
              points([2,3,1],:), points([3,1,2],:), points([3,2,1],:)];

    points5 = points/5*3;
    points5(3,:) = points5(3,:) + 0.8;

    points = [points, points5];
    
    points = [points; zeros(size(points)); 255*ones(size(points))]; 
    points(8:9,:) = 0;
end

function points = light_visualization(  )
    % creates a point cloud visualization of a light source
    
    [X,Y,Z] = meshgrid(-0.5:0.1:0.5,-0.5:0.1:0.5,-0.5:0.1:0.5);
    normXYZ = sqrt(X.^2+Y.^2+Z.^2);
    normXYZ(normXYZ==0) = 1;

    points = [X(:),Y(:),Z(:)]'./repmat(normXYZ(:),1,3)';

    points = [points; zeros(size(points)); 255*ones(size(points))];
    points(9,:) = 0;
end

function savePLY( dest, points )
    % writes binary ply-file from point cloud array

    N = size(points, 2);
    fileID = fopen(dest, 'w');
    fprintf(fileID, 'ply\n');
    fprintf(fileID, 'format binary_little_endian 1.0\n');
    fprintf(fileID, ['element vertex ', num2str(N), '\n']);
    fprintf(fileID, 'property float x\n');
    fprintf(fileID, 'property float y\n');
    fprintf(fileID, 'property float z\n');
    fprintf(fileID, 'property float nx\n');
    fprintf(fileID, 'property float ny\n');
    fprintf(fileID, 'property float nz\n');
    fprintf(fileID, 'property uchar red\n');
    fprintf(fileID, 'property uchar green\n');
    fprintf(fileID, 'property uchar blue\n');
    fprintf(fileID, 'end_header\n');

    colors = 255 * ones(3, N);
    normals = zeros(3, N);
    if size(points, 1) == 6
        colors = points(4:6,:);
    end
    if size(points, 1) == 9
        normals = points(4:6,:);
        colors = points(7:9,:);
    end

    points_8bit = reshape(typecast(single(reshape(points(1:3,:), 3*N, 1)), 'uint8'), 12, N);
    normals_8bit = reshape(typecast(single(reshape(normals, 3*N, 1)), 'uint8'), 12, N);
    combined = reshape([points_8bit; normals_8bit; colors], 27*N, 1);

    fwrite(fileID, combined, 'uint8');
    fclose(fileID);
end

function rgb_flow = flow2rgb(flowX, flowY, maxFlow)
    % creates flow field using x- and y-displacements and maximal flow
  
    mask = (flowX~=0).*(flowY~=0).*(~isnan(flowX)).*(~isnan(flowY));
    
    rgb_flow = cat(3, flowX, -(flowX+flowY)/2, flowY)/maxFlow + 1;
    rgb_flow = mask .* rgb_flow;
    
    rgb_flow(rgb_flow<0) = 0;
    rgb_flow(rgb_flow>1) = 1;
end

function img1_w = warp(img1, mask0, flowX0, flowY0)
    % warps image1 towards image0 using backward flow0
    
    [h, w, ~] = size(img1);
    [X, Y] = meshgrid(1:w, 1:h);
    posX = min(max(round(X + flowX0), 1), w);
    posY = min(max(round(Y + flowY0), 1), h);
    linIdx = sub2ind([h, w], posY(:), posX(:));
    img1_w = uint8(mask0).*reshape(img1(linIdx), h, w);
    img1_w = cat(3, img1_w, img1_w, img1_w);
end